from hseb.core.config import (
    QuantDatatype,
)

from tests.engine.base import EngineParams, EngineSuite


class TestWeaviateEngine(EngineSuite):
    def params(self):
        return EngineParams(
            engine="hseb.engine.weaviate.WeaviateEngine",
            image="semitechnologies/weaviate:1.32.8",
            quantizations=[QuantDatatype.INT8, QuantDatatype.INT1],
        )



from hseb.core.config import IndexArgs, QuantDatatype, SearchArgs
from hseb.core.measurement import ExperimentResult, QueryResult
from hseb.core.response import DocScore
import tempfile
from pathlib import Path


def test_write_read():
    m = QueryResult(query_id=1, exact=[DocScore(1, 1)], response=[DocScore(1, 1)], client_latency=1)
    result = ExperimentResult(
        tag="test",
        index_args=IndexArgs(
            m=32,
            ef_construction=32,
            quant=QuantDatatype.FLOAT32,
            batch_size=32,
            segments=None,
            kwargs={},
        ),
        search_args=SearchArgs(ef_search=32, filter_selectivity=100, kwargs={}),
        queries=[m],
        indexing_time=[1],
        warmup_latencies=[1],
    )
    with tempfile.TemporaryDirectory(prefix="hseb_test_") as dir:
        result.to_json(dir)
        for file in Path(dir).iterdir():
            decoded = ExperimentResult.from_json(file)
            assert decoded == result



from __future__ import annotations
import yaml
from typing import Any
from itertools import product
from pydantic import BaseModel, Field
from dataclasses import dataclass, asdict
from structlog import get_logger
from enum import StrEnum

logger = get_logger()


class Config(BaseModel):
    engine: str = Field(min_length=1, description="engine type")
    image: str = Field(min_length=1, description="docker image to run")
    dataset: DatasetConfig
    experiments: list[ExperimentConfig]

    @staticmethod
    def from_yaml(text: str) -> Config:
        parsed: dict = yaml.safe_load(text)
        return Config(**parsed)

    @staticmethod
    def from_file(path: str) -> Config:
        with open(path, "r") as f:
            config = Config.from_yaml(f.read())
            logger.info(f"Loaded config file from {path}, engine: {config.engine}")
            return config


class DatasetConfig(BaseModel):
    dim: int = Field(gt=0)
    name: str
    query: str = Field(description="queries dataset name")
    corpus: str = Field(description="corpus dataset name")


class ExperimentConfig(BaseModel):
    tag: str
    k: int
    index: IndexArgsMatrix
    search: SearchArgsMatrix


@dataclass
class IndexArgs:
    m: int
    ef_construction: int
    quant: QuantDatatype
    batch_size: int
    segments: int | None  # none means default
    kwargs: dict[str, Any]

    def to_string(self) -> str:
        parts1 = [f"{key}={value}" for key, value in asdict(self).items() if key != "kwargs"]
        parts2 = [f"{key}={value}" for key, value in self.kwargs.items()]
        return "_".join(parts1 + parts2)


class IndexArgsMatrix(BaseModel):
    m: list[int]
    ef_construction: list[int]
    quant: list[QuantDatatype]
    batch_size: int = 1024
    segments: list[int] | None = Field(default=None)
    kwargs: dict[str, list] = Field(default_factory=dict)

    def expand(self) -> list[IndexArgs]:
        """Generate all permutations of parameters from IndexArgs."""
        segments_list = [None] if self.segments is None or len(self.segments) == 0 else self.segments
        base_params = product(self.m, self.ef_construction, self.quant, segments_list)
        kwargs_combos = product(*self.kwargs.values()) if self.kwargs else [()]

        return [
            IndexArgs(
                m=m,
                ef_construction=ef,
                quant=quant,
                batch_size=self.batch_size,
                segments=segments,
                kwargs=dict(zip(self.kwargs.keys(), kwarg_vals)),
            )
            for (m, ef, quant, segments), kwarg_vals in product(base_params, kwargs_combos)
        ]


class QuantDatatype(StrEnum):
    FLOAT32 = "float32"
    FLOAT16 = "float16"
    INT8 = "int8"
    INT1 = "int1"


@dataclass
class SearchArgs:
    ef_search: int
    filter_selectivity: int
    kwargs: dict[str, Any]

    def to_string(self) -> str:
        parts1 = [f"{key}={value}" for key, value in asdict(self).items() if key != "kwargs"]
        parts2 = [f"{key}={value}" for key, value in self.kwargs.items()]
        return "_".join(parts1 + parts2)


class SearchArgsMatrix(BaseModel):
    ef_search: list[int]
    filter_selectivity: list[int]
    kwargs: dict[str, list] = Field(default_factory=dict)

    def expand(self) -> list[SearchArgs]:
        """Generate all permutations of parameters from IndexArgs."""
        base_params = product(self.ef_search, self.filter_selectivity)
        kwargs_combos = product(*self.kwargs.values()) if self.kwargs else [()]

        return [
            SearchArgs(
                ef_search=ef_search,
                filter_selectivity=filter_selectivity,
                kwargs=dict(zip(self.kwargs.keys(), kwarg_vals)),
            )
            for (ef_search, filter_selectivity), kwarg_vals in product(base_params, kwargs_combos)
        ]



from hseb.core.config import IndexArgs, QuantDatatype, SearchArgs
from hseb.core.measurement import ExperimentResult, QueryResult
from hseb.core.report import Report
from hseb.core.response import DocScore
from hseb.core.submission import ExperimentMetrics


def test_report():
    m = QueryResult(query_id=1, exact=[DocScore(1, 1)], response=[DocScore(1, 1)], client_latency=1)
    result = ExperimentResult(
        tag="test",
        index_args=IndexArgs(
            m=32,
            ef_construction=32,
            quant=QuantDatatype.FLOAT32,
            batch_size=32,
            segments=None,
            kwargs={},
        ),
        search_args=SearchArgs(ef_search=32, filter_selectivity=100, kwargs={}),
        queries=[m],
        indexing_time=[1],
        warmup_latencies=[1],
    )
    submission = Report.from_experiments([ExperimentMetrics.from_experiment(result)])
    assert submission.df.shape == (1, 18)



from __future__ import annotations
import pandas as pd
import numpy as np
from dataclasses import dataclass
from collections import defaultdict
from hseb.core.submission import ExperimentMetrics
from tqdm import tqdm


@dataclass
class Report:
    df: pd.DataFrame

    @staticmethod
    def from_experiments(experiments: list[ExperimentMetrics]) -> Report:
        columns: dict[str, list] = defaultdict(list)
        for exp in tqdm(experiments, desc="processing experiments"):
            columns["tag"].append(exp.tag)
            columns["indexing_time_total"].append(sum(exp.indexing_time))
            columns["indexing_time_batch_mean"].append(sum(exp.indexing_time) / len(exp.indexing_time))
            columns["indexing_time_batch_median"].append(np.percentile(exp.indexing_time, 50).item())
            columns["ef_construction"].append(exp.index_args.ef_construction)
            columns["m"].append(exp.index_args.m)
            columns["batch_size"].append(exp.index_args.batch_size)
            columns["segments"].append(exp.index_args.segments)
            for key, value in exp.index_args.kwargs.items():
                columns[key].append(value)
            columns["ef_search"].append(exp.search_args.ef_search)
            columns["filter_selectivity"].append(exp.search_args.filter_selectivity)
            for key, value in exp.search_args.kwargs.items():
                columns[key].append(value)
            for metric, value in exp.metrics.as_dict().items():
                columns[metric].append(value)
        return Report(pd.DataFrame(columns))



from hseb.core.config import (
    QuantDatatype,
)

from tests.engine.base import EngineParams, EngineSuite


class TestRedisEngine(EngineSuite):
    def params(self):
        return EngineParams(
            engine="hseb.engine.redis.RedisEngine",
            image="redis:8.2.1",
            quantizations=[QuantDatatype.FLOAT16],
        )



import logging
import time

import docker
import psycopg2
from psycopg2.extras import execute_values

from hseb.core.config import Config, IndexArgs, QuantDatatype, SearchArgs
from hseb.core.dataset import Doc, Query
from hseb.core.response import DocScore, IndexResponse, SearchResponse
from hseb.engine.base import EngineBase

logger = logging.getLogger()

POSTGRES_DATATYPES = {
    QuantDatatype.FLOAT32: "vector",
    QuantDatatype.FLOAT16: "halfvec",
    QuantDatatype.INT1: "bit",
}

POSTGRES_INDEX_OPS = {
    QuantDatatype.FLOAT32: "vector_ip_ops",
    QuantDatatype.FLOAT16: "halfvec_ip_ops",
    QuantDatatype.INT1: "bit_hamming_ops",
}


class PostgresEngine(EngineBase):
    def __init__(self, config: Config):
        self.config = config
        self.connection = None
        self.container = None

    def start(self, index_args: IndexArgs) -> None:
        # Check for unsupported quantization types
        if index_args.quant == QuantDatatype.INT8:
            raise ValueError("Postgres with pgvector does not support INT8 quantization")
        if index_args.segments is not None:
            raise ValueError("Postgres with pgvector cannot set number of segments")
        if index_args.ef_construction < index_args.m * 2:
            raise ValueError(
                f"pgvector requires efc >= 2*m (current ef={index_args.ef_construction} m={index_args.m})"
            )

        docker_client = docker.from_env()

        # PostgreSQL configuration for better performance
        shared_buffers = index_args.kwargs.get("shared_buffers", "2GB")
        work_mem = index_args.kwargs.get("work_mem", "16MB")
        maintenance_work_mem = index_args.kwargs.get("maintenance_work_mem", "512MB")

        self.container = docker_client.containers.run(
            image=self.config.image,
            ports={"5432/tcp": 5432},
            detach=True,
            shm_size="4g",  # needed for buffers
            environment={
                "POSTGRES_DB": "benchmark",
                "POSTGRES_USER": "postgres",
                "POSTGRES_PASSWORD": "postgres",
                "POSTGRES_HOST_AUTH_METHOD": "trust",
            },
            command=[
                "postgres",
                "-c",
                f"shared_buffers={shared_buffers}",
                "-c",
                f"work_mem={work_mem}",
                "-c",
                f"maintenance_work_mem={maintenance_work_mem}",
                "-c",
                "max_connections=100",
                "-c",
                "random_page_cost=1.1",
                "-c",
                "effective_cache_size=1GB",
                "-c",
                "max_parallel_maintenance_workers=7",
            ],
        )
        self._wait_for_logs(self.container, "database system is ready to accept connections")

        # Wait a bit more to ensure PostgreSQL is fully ready
        time.sleep(2)

        # Connect to database
        self.connection = psycopg2.connect(
            host="localhost", port=5432, database="benchmark", user="postgres", password="postgres"
        )
        self.connection.autocommit = True

        # Create pgvector extension and table
        with self.connection.cursor() as cursor:
            cursor.execute("CREATE EXTENSION IF NOT EXISTS vector;")

            datatype = POSTGRES_DATATYPES[index_args.quant]
            cursor.execute(f"""
                CREATE UNLOGGED TABLE IF NOT EXISTS documents (
                    id INTEGER PRIMARY KEY,
                    embedding {datatype}({self.config.dataset.dim}),
                    tag INTEGER[]
                );
            """)
            cursor.execute("SET max_parallel_maintenance_workers = 7;")
            cursor.execute("SET client_min_messages = DEBUG;")

            time.sleep(5)
            # Create HNSW index with appropriate operator class
            ops_class = POSTGRES_INDEX_OPS[index_args.quant]
            cursor.execute(
                f"""
                CREATE INDEX IF NOT EXISTS documents_embedding_idx 
                ON documents USING hnsw (embedding {ops_class})
                WITH (m = %s, ef_construction = %s);
            """,
                (index_args.m, index_args.ef_construction),
            )

            # Create GIN index on tag array for efficient filtering
            cursor.execute("CREATE INDEX IF NOT EXISTS documents_tag_idx ON documents USING gin (tag);")
            time.sleep(5)

        self.index_args = index_args  # Store for later use
        return self

    def stop(self, cleanup: bool):
        if self.connection:
            self.connection.close()
        if self.container:
            self.container.stop()
        if cleanup and self.container:
            self.container.remove(v=True)

    def commit(self):
        # Just optimize table
        with self.connection.cursor() as cursor:
            cursor.execute("VACUUM ANALYZE documents;")

    def index_is_green(self) -> bool:
        with self.connection.cursor() as cursor:
            cursor.execute("""SELECT phase, blocks_done, blocks_total FROM pg_stat_progress_create_index;""")
            status = cursor.fetchall()
            return len(status) == 0

    def index_batch(self, batch: list[Doc]) -> IndexResponse:
        # Prepare data for batch insert
        data = [(doc.id, doc.embedding.tolist(), doc.tag) for doc in batch]

        with self.connection.cursor() as cursor:
            start = time.perf_counter()
            execute_values(
                cursor,
                "INSERT INTO documents (id, embedding, tag) VALUES %s ON CONFLICT (id) DO NOTHING",
                data,
                template=None,
                page_size=len(data),
            )
            end = time.perf_counter()
            return IndexResponse(client_latency=end - start)

    def search(self, search_params: SearchArgs, query: Query, top_k: int) -> SearchResponse:
        # Build the query - use appropriate distance function with proper type casting
        if self.index_args.quant == QuantDatatype.INT1:
            # For bit vectors, use Hamming distance
            sql = """
                SELECT id, embedding <~> %s::bit AS distance
                FROM documents
            """
        elif self.index_args.quant == QuantDatatype.FLOAT16:
            # For halfvec, cast to halfvec type
            sql = """
                SELECT id, embedding <#> %s::halfvec AS distance  
                FROM documents
            """
        else:
            # For float32 vectors, cast to vector type
            sql = """
                SELECT id, embedding <#> %s::vector AS distance  
                FROM documents
            """

        params = [query.embedding.tolist()]

        # Add filtering if needed
        if search_params.filter_selectivity != 100:
            sql += " WHERE %s = ANY(tag)"
            params.append(search_params.filter_selectivity)

        sql += " ORDER BY distance LIMIT %s"
        params.append(top_k)

        with self.connection.cursor() as cursor:
            cursor.execute("SET hnsw.ef_search = %s", (search_params.ef_search,))
            cursor.execute("SET hnsw.iterative_scan = strict_order")
            start = time.time_ns()
            cursor.execute(sql, params)
            results = cursor.fetchall()
        end = time.time_ns()

        # Convert to DocScore objects
        if self.index_args.quant == QuantDatatype.INT1:
            # For Hamming distance, convert to similarity score (lower distance = higher similarity)
            doc_scores = [DocScore(doc=row[0], score=1.0 / (1.0 + row[1])) for row in results]
        else:
            # For inner product, convert negative inner product to positive score
            doc_scores = [DocScore(doc=row[0], score=-row[1]) for row in results]

        return SearchResponse(
            results=doc_scores,
            client_latency=(end - start) / 1000000000.0,
        )



from hseb.core.config import (
    QuantDatatype,
)

from tests.engine.base import EngineParams, EngineSuite


class TestElasticengine(EngineSuite):
    def params(self):
        return EngineParams(
            engine="hseb.engine.elastic.ElasticsearchEngine",
            image="elasticsearch:9.1.3",
            quantizations=[QuantDatatype.INT8, QuantDatatype.INT1],
        )



from __future__ import annotations
import datetime
import json
from pydantic import BaseModel, Field
from hseb.core.config import Config, IndexArgs, SearchArgs
from tqdm import tqdm
from pathlib import Path
from importlib.metadata import PackageNotFoundError, version
from hseb.core.measurement import ExperimentResult
from structlog import get_logger
import numpy as np

logger = get_logger()


class QueryMetrics(BaseModel):
    recall5: list[float] = Field(default_factory=list)
    recall10: list[float] = Field(default_factory=list)
    recall30: list[float] = Field(default_factory=list)
    recall100: list[float] = Field(default_factory=list)
    latency: list[float] = Field(default_factory=list)

    def as_dict(self) -> dict[str, float]:
        return {
            "recall@5": sum(self.recall5) / len(self.recall5),
            "recall@10": sum(self.recall10) / len(self.recall10),
            "recall@30": sum(self.recall30) / len(self.recall30),
            "recall@100": sum(self.recall100) / len(self.recall100),
            "latency_mean": sum(self.latency) / len(self.latency),
            "latency_p90": np.percentile(self.latency, q=90).item(),
            "latency_p95": np.percentile(self.latency, q=95).item(),
            "latency_p99": np.percentile(self.latency, q=99).item(),
        }

    def as_string(self) -> str:
        values = [f"{key}={round(value, 5)}" for key, value in self.as_dict().items()]
        return " ".join(values)


class ExperimentMetrics(BaseModel):
    tag: str
    indexing_time: list[float]
    index_args: IndexArgs
    search_args: SearchArgs
    metrics: QueryMetrics
    warmup_latencies: list[float]

    @staticmethod
    def recall_score(true_docs: list[int], retrieved_docs: list[int], k: int) -> float:
        k_adj = min(k, len(retrieved_docs), len(true_docs))
        retrieved_set = set(retrieved_docs[:k_adj])
        true_set = set(true_docs[:k_adj])
        hits = len(retrieved_set.intersection(true_set))
        return hits / float(len(true_set))

    @staticmethod
    def from_experiment(exp: ExperimentResult) -> ExperimentMetrics:
        metrics = QueryMetrics()
        for query in exp.queries:
            ground_truth = [doc.doc for doc in query.exact]
            retrieved_docs = [doc.doc for doc in query.response]
            metrics.recall5.append(ExperimentMetrics.recall_score(ground_truth, retrieved_docs, 5))
            metrics.recall10.append(ExperimentMetrics.recall_score(ground_truth, retrieved_docs, 10))
            metrics.recall30.append(ExperimentMetrics.recall_score(ground_truth, retrieved_docs, 30))
            metrics.recall100.append(ExperimentMetrics.recall_score(ground_truth, retrieved_docs, 100))
            metrics.latency.append(query.client_latency)
        return ExperimentMetrics(
            tag=exp.tag,
            indexing_time=exp.indexing_time,
            index_args=exp.index_args,
            search_args=exp.search_args,
            metrics=metrics,
            warmup_latencies=exp.warmup_latencies,
        )


class Submission(BaseModel):
    time: str
    version: str
    config: Config
    experiments: list[ExperimentMetrics]

    @staticmethod
    def from_dir(config: Config, path: str) -> Submission:
        experiments = []
        for file in tqdm(list(Path(path).iterdir()), desc="loading measurements"):
            if file.is_file() and file.name.endswith(".json"):
                exp = ExperimentResult.from_json(file)
                experiments.append(ExperimentMetrics.from_experiment(exp))
        try:
            hseb_version = version("hseb")
        except PackageNotFoundError:
            hseb_version = "unknown"
        logger.info(f"Loaded {len(experiments)} experiments")
        return Submission(
            time=datetime.datetime.now().isoformat(),
            version=hseb_version,
            config=config,
            experiments=experiments,
        )

    @staticmethod
    def from_json(path: str) -> Submission:
        with open(path, "r") as file:
            raw = json.loads(file.read())
            return Submission(**raw)

    def to_json(self, path: str):
        if Path(path).exists():
            raise Exception(f"output file {path} already exists")
        with open(path, "w") as file:
            file.write(json.dumps(self.model_dump()))



import docker
from hseb.core.config import Config, IndexArgs, QuantDatatype, SearchArgs
from hseb.core.dataset import Doc, Query
from hseb.core.response import DocScore, IndexResponse, SearchResponse
from hseb.engine.base import EngineBase
from qdrant_client import QdrantClient
import time
from qdrant_client.models import (
    VectorParams,
    Distance,
    PointStruct,
    ScalarQuantizationConfig,
    ScalarType,
    BinaryQuantizationConfig,
    Filter,
    FieldCondition,
    MatchValue,
    SearchParams,
    CollectionStatus,
    OptimizersConfigDiff,
    HnswConfigDiff,
    ScalarQuantization,
    BinaryQuantization,
)

QDRANT_DATATYPES = {
    QuantDatatype.INT8: ScalarQuantization(scalar=ScalarQuantizationConfig(type=ScalarType.INT8)),
    QuantDatatype.INT1: BinaryQuantization(binary=BinaryQuantizationConfig()),
}


class Qdrant(EngineBase):
    def __init__(self, config: Config):
        self.config = config

    def index_batch(self, batch: list[Doc]) -> IndexResponse:
        points = [PointStruct(id=doc.id, vector=doc.embedding.tolist(), payload={"tag": doc.tag}) for doc in batch]
        start = time.perf_counter()
        self.client.upsert(collection_name="test", points=points)
        end = time.perf_counter()
        return IndexResponse(client_latency=end - start)

    def commit(self):
        pass

    def index_is_green(self) -> bool:
        status = self.client.get_collection(collection_name="test")
        return status.status == CollectionStatus.GREEN

    def search(self, search_params: SearchArgs, query: Query, top_k: int) -> SearchResponse:
        def create_filter():
            if search_params.filter_selectivity == 100:
                return None
            else:
                return Filter(must=FieldCondition(key="tag", match=MatchValue(value=search_params.filter_selectivity)))

        start = time.time_ns()
        response = self.client.query_points(
            collection_name="test",
            query=query.embedding,
            query_filter=create_filter(),
            search_params=SearchParams(
                hnsw_ef=search_params.ef_search,
                indexed_only=True,  # to avoid fullscans over memtable
                exact=False,
            ),
            limit=top_k,
        )
        end = time.time_ns()
        return SearchResponse(
            results=[DocScore(point.id, point.score) for point in response.points],
            client_latency=(end - start) / 1000000000.0,
        )

    def start(self, index_args: IndexArgs):
        docker_client = docker.from_env()
        self.container = docker_client.containers.run(
            image=self.config.image,
            ports={"6333/tcp": 6333},
            detach=True,
        )
        self._wait_for_logs(self.container, "Qdrant gRPC listening")

        self.client = QdrantClient(host="localhost")
        self.client.create_collection(
            collection_name="test",
            vectors_config=VectorParams(
                size=self.config.dataset.dim,
                distance=Distance.DOT,
                on_disk=index_args.kwargs.get("original_vectors_on_disk", None),
                datatype=None,  # it's about only the raw vectors, not the index
            ),
            quantization_config=QDRANT_DATATYPES.get(index_args.quant, None),
            hnsw_config=HnswConfigDiff(
                m=index_args.m,
                ef_construct=index_args.ef_construction,
                on_disk=index_args.kwargs.get("hnsw_on_disk", None),
            ),
            optimizers_config=OptimizersConfigDiff(
                max_segment_size=index_args.kwargs.get("max_segment_size_kb", None),
                default_segment_number=index_args.kwargs.get("default_segment_number", index_args.segments),
            ),
        )
        self.client.create_payload_index(collection_name="test", field_name="tag", field_schema="integer")
        return self

    def stop(self, cleanup: bool):
        self.container.stop()
        if cleanup and self.container:
            self.container.remove(v=True)



import argparse
import time
from hseb.core.submission import ExperimentMetrics, Submission
from hseb.engine.base import EngineBase
from hseb.core.config import Config
from hseb.core.dataset import BenchmarkDataset
from hseb.core.measurement import ExperimentResult, QueryResult
from tqdm import tqdm
from structlog import get_logger
import tempfile
import random

logger = get_logger()

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        required=True,
        type=str,
        help="path to a config file",
    )
    parser.add_argument(
        "--out",
        type=str,
        required=True,
        help="output file name",
    )
    parser.add_argument(
        "--cleanup",
        type=bool,
        required=False,
        default=True,
        help="should we delete all stopped containers? this saves disk space, but you lose engine logs",
    )
    parser.add_argument(
        "--warmup",
        type=int,
        required=False,
        default=1000,
        help="number of random queries to warmup the engine",
    )
    parser.add_argument(
        "--queries",
        type=int,
        required=False,
        help="number of queries to sample for benchmark",
    )
    parser.add_argument(
        "--index-wait-seconds",
        type=int,
        required=False,
        default=300,
        help="how long should we wait for an index to become green",
    )

    args = parser.parse_args()

    config = Config.from_file(args.config)
    data = BenchmarkDataset(config.dataset)
    engine = EngineBase.load_class(config.engine, config)
    with tempfile.TemporaryDirectory(prefix="hseb_", delete=args.cleanup) as workdir:
        logger.info(f"Initialized engine, workdir: {workdir}")
        run_index = 0
        index_fails = 0
        search_fails = 0
        search_request_fails = 0
        for exp_index, exp in enumerate(config.experiments):
            index_variations = exp.index.expand()
            search_variations = exp.search.expand()
            total_cases = len(index_variations) * len(search_variations)
            logger.info(
                f"Running experiment {exp_index + 1} of {len(config.experiments)}. Targets: index={len(index_variations)} search={len(search_variations)} total={total_cases}"
            )

            for indexing_args_index, index_args in enumerate(index_variations):
                logger.info(f"Indexing run {indexing_args_index + 1}/{len(index_variations)}: {index_args}")
                try:
                    engine.start(index_args)
                    batches = data.corpus_batched(index_args.batch_size)
                    total = int(len(data.corpus_dataset) / index_args.batch_size)
                    index_start = time.perf_counter()
                    index_batch_times: list[float] = []
                    for batch in tqdm(batches, total=total, desc="indexing"):
                        index_response = engine.index_batch(batch=batch)
                        index_batch_times.append(index_response.client_latency)

                    commit_start = time.perf_counter()
                    engine.commit()
                    poll_green_attempts = 0
                    is_green = False
                    while not is_green and poll_green_attempts < args.index_wait_seconds:
                        is_green = engine.index_is_green()
                        poll_green_attempts += 1
                        if not is_green:
                            time.sleep(1)
                    if not is_green:
                        raise Exception(f"Index stuck in non-green status for {args.index_wait_seconds} seconds")
                    try:
                        warmup_start = time.perf_counter()
                        logger.info(
                            f"Index built in {warmup_start - index_start} seconds (ingest={int(commit_start - index_start)} commit={int(warmup_start - commit_start)})"
                        )
                        warmup_latencies: list[float] = []
                        for warmup_query in tqdm(
                            random.choices(list(data.queries(limit=1000)), k=args.warmup), desc="warmup"
                        ):
                            k_eff = min(exp.k, search_variations[0].ef_search)
                            response = engine.search(search_variations[0], warmup_query, k_eff)
                            warmup_latencies.append(response.client_latency)

                        logger.info(f"Warmup done in {time.perf_counter() - warmup_start} seconds")
                        for search_args_index, search_args in enumerate(search_variations):
                            logger.info(
                                f"Search {search_args_index + 1}/{len(search_variations)} ({run_index + 1}/{total_cases}): {search_args}"
                            )

                            measurements: list[QueryResult] = []
                            k_eff = min(exp.k, search_args.ef_search)
                            incomplete_results: list[int] = []
                            failed_queries = 0
                            for query in tqdm(list(data.queries(limit=args.queries)), desc="search"):
                                try:
                                    response = engine.search(search_args, query, k_eff)
                                    if len(response.results) != k_eff:
                                        incomplete_results.append(len(response.results))
                                    measurements.append(
                                        QueryResult.from_response(
                                            query=query, search_args=search_args, response=response
                                        )
                                    )
                                except Exception as error:
                                    logger.exception(f"got search request error {error}")
                                    failed_queries += 1
                                    search_request_fails += 1

                            result = ExperimentResult(
                                tag=exp.tag,
                                index_args=index_args,
                                search_args=search_args,
                                queries=measurements,
                                indexing_time=index_batch_times,
                                warmup_latencies=warmup_latencies,
                            )
                            result.to_json(workdir=workdir)
                            metrics = ExperimentMetrics.from_experiment(result)
                            if len(incomplete_results) > 0:
                                logger.warning(
                                    f"Engine returned {len(incomplete_results)}/{len(measurements)} incomplete results (expected {k_eff}): {incomplete_results}"
                                )
                            if failed_queries > 0:
                                logger.warning(f"Engine got {failed_queries} queries with errors")
                            logger.info(f"Run finished: {metrics.metrics.as_string()}")
                            run_index += 1
                    except Exception as error:
                        logger.exception(f"skipping search run {search_args}: {error}")
                        search_fails += 1
                    logger.debug(
                        f"Indexing run {indexing_args_index + 1}/{len(index_variations)} done in {time.perf_counter() - index_start} seconds"
                    )
                except Exception as error:
                    logger.exception(f"skipping index run {index_args}: {error}")
                    index_fails += 1
                finally:
                    engine.stop(args.cleanup)
        submission = Submission.from_dir(config=config, path=workdir)
        logger.info(f"Writing submission file to {args.out}")
        submission.to_json(args.out)
        logger.info(
            f"Experiment finished. {index_fails} index fails, {search_fails} search fails ({search_request_fails} failed requests)."
        )



import logging
import time

import docker
from elasticsearch import Elasticsearch, helpers

from hseb.core.config import Config, IndexArgs, QuantDatatype, SearchArgs
from hseb.core.dataset import Doc, Query
from hseb.core.response import DocScore, IndexResponse, SearchResponse
from hseb.engine.base import EngineBase

logger = logging.getLogger()

ES_HNSW_TYPES = {
    QuantDatatype.FLOAT32: "hnsw",
    QuantDatatype.INT8: "int8_hnsw",
    QuantDatatype.INT1: "bbq_hnsw",
}


class ElasticsearchEngine(EngineBase):
    def __init__(self, config: Config):
        self.config = config

    def start(self, index_args: IndexArgs):
        self.index_args = index_args
        docker_client = docker.from_env()
        heap_size = index_args.kwargs.get("heap_size", "8g")
        self.container = docker_client.containers.run(
            image=self.config.image,
            ports={"9200/tcp": 9200, "9300/tcp": 9300},
            detach=True,
            environment={
                "discovery.type": "single-node",
                "ES_JAVA_OPTS": f"-Xms{heap_size} -Xmx{heap_size}",
                "xpack.security.enabled": "false",
            },
        )
        self._wait_for_logs(self.container, "is selected as the current health node")
        self.client = Elasticsearch("http://localhost:9200", request_timeout=30)
        # we control segment size by calling refresh
        settings = {
            "index": {
                "refresh_interval": "1h",
                "number_of_replicas": 0,
                "number_of_shards": 1,
            }
        }
        if "max_merged_segment" in index_args.kwargs:
            settings["merge"] = {"policy": {"max_merged_segment": index_args.kwargs["max_merged_segment"]}}

        self.client.indices.create(
            index="test",
            settings=settings,
            mappings={
                "properties": {
                    # "_id": {"type": "integer"},
                    "text": {
                        "type": "dense_vector",
                        "dims": self.config.dataset.dim,
                        "index": True,
                        "similarity": "dot_product",
                        "index_options": {
                            # use the same as element_type to make it simpler - we're not measuring rescoring yet
                            "type": ES_HNSW_TYPES[index_args.quant],
                            "m": index_args.m,
                            "ef_construction": index_args.ef_construction,
                        },
                    },
                    "tag": {"type": "integer"},
                }
            },
        )
        self.docs_in_segment = 0

    def stop(self, cleanup: bool):
        self.container.stop()
        if cleanup and self.container:
            self.container.remove(v=True)

    def commit(self):
        self.client.indices.refresh(index="test")
        if self.index_args.segments is not None:
            self.client.indices.forcemerge(
                index="test",
                max_num_segments=self.index_args.segments,
                wait_for_completion=True,
            )

    def index_is_green(self) -> bool:
        response = self.client.cluster.health(index="test")
        return response["status"] == "green"

    def index_batch(self, batch: list[Doc]) -> IndexResponse:
        actions = []
        for doc in batch:
            actions.append(
                {
                    "_op_type": "index",
                    "_index": "test",
                    "_source": {
                        "text": doc.embedding.tolist(),
                        "tag": doc.tag,
                    },
                    "_id": doc.id,
                }
            )
        start = time.perf_counter()
        helpers.bulk(self.client, actions)
        end = time.perf_counter()
        self.docs_in_segment += len(batch)
        if self.docs_in_segment >= self.index_args.kwargs.get("docs_per_segment", 1024):
            self.client.indices.refresh(index="test")
            self.docs_in_segment = 0
        return IndexResponse(client_latency=end - start)

    def search(self, search_params: SearchArgs, query: Query, top_k: int) -> SearchResponse:
        es_query = {
            "field": "text",
            "query_vector": query.embedding.tolist(),
            "k": top_k,
            "num_candidates": search_params.ef_search,
        }
        if search_params.filter_selectivity != 100:
            es_query["filter"] = {"terms": {"tag": [search_params.filter_selectivity]}}
        start = time.time_ns()
        response = self.client.search(index="test", knn=es_query, source=["_id"], size=top_k)
        end = time.time_ns()
        return SearchResponse(
            results=[DocScore(doc=int(doc["_id"]), score=doc["_score"]) for doc in response["hits"]["hits"]],
            client_latency=(end - start) / 1000000000.0,
        )



from hseb.core.config import (
    QuantDatatype,
)

from tests.engine.base import EngineParams, EngineSuite


class TestNixiesearch(EngineSuite):
    def params(self):
        return EngineParams(
            engine="hseb.engine.nixiesearch.NixiesearchEngine",
            image="nixiesearch/nixiesearch:0.7.2",
            quantizations=[QuantDatatype.INT8, QuantDatatype.INT1],
        )



from dataclasses import dataclass


@dataclass
class DocScore:
    doc: int
    score: float


@dataclass
class SearchResponse:
    results: list[DocScore]
    client_latency: float


@dataclass
class IndexResponse:
    client_latency: float



from __future__ import annotations
from pydantic import BaseModel
from hseb.core.dataset import Query
from hseb.core.response import DocScore, SearchResponse
from hseb.core.config import IndexArgs, SearchArgs

import json
from structlog import get_logger

logger = get_logger()


class ExperimentResult(BaseModel):
    tag: str
    indexing_time: list[float]
    index_args: IndexArgs
    search_args: SearchArgs
    queries: list[QueryResult]
    warmup_latencies: list[float]

    @staticmethod
    def from_json(path: str) -> ExperimentResult:
        with open(path, "r") as file:
            raw = json.load(file)
            return ExperimentResult(**raw)

    def to_json(self, workdir: str):
        out_file = f"{workdir}/{self.tag}-{self.index_args.to_string()}-{self.search_args.to_string()}.json"
        # logger.debug(f"Saved experiment result to {out_file}")
        with open(out_file, "w") as file:
            file.write(json.dumps(self.model_dump()))


class QueryResult(BaseModel):
    query_id: int
    exact: list[DocScore]
    response: list[DocScore]
    client_latency: float

    @staticmethod
    def from_response(
        query: Query,
        search_args: SearchArgs,
        response: SearchResponse,
    ) -> QueryResult:
        match search_args.filter_selectivity:
            case 10:
                exact = query.exact10
            case 90:
                exact = query.exact90
            case 100:
                exact = query.exact100
        return QueryResult(
            query_id=query.id,
            exact=exact,
            response=response.results,
            client_latency=response.client_latency,
        )



from __future__ import annotations
from hseb.core.config import DatasetConfig
from datasets import load_dataset, Features, Value, Sequence, Dataset
from dataclasses import dataclass
import numpy as np
from typing import Generator, Any
from hseb.core.response import DocScore
from structlog import get_logger

logger = get_logger()

CORPUS_SCHEMA = Features(
    {
        "id": Value("int32"),
        "text": Value("string"),
        "embedding": Sequence(Value("float32")),
        "tag": Sequence(Value("int32")),
    }
)

QUERY_SCHEMA = Features(
    {
        "id": Value("int32"),
        "text": Value("string"),
        "embedding": Sequence(Value("float32")),
        "results_10_docs": Sequence(Value("int32")),
        "results_10_scores": Sequence(Value("float32")),
        "results_90_docs": Sequence(Value("int32")),
        "results_90_scores": Sequence(Value("float32")),
        "results_100_docs": Sequence(Value("int32")),
        "results_100_scores": Sequence(Value("float32")),
    }
)


class BenchmarkDataset:
    query_dataset: Dataset
    corpus_dataset: Dataset

    def __init__(self, config: DatasetConfig) -> None:
        self.query_dataset = load_dataset(config.name, config.query, split="train", features=QUERY_SCHEMA)
        self.corpus_dataset = load_dataset(config.name, config.corpus, split="train", features=CORPUS_SCHEMA)
        logger.info(f"Loaded dataset: {len(self.query_dataset)} queries, {len(self.corpus_dataset)} documents")

    def corpus_batched(self, batch_size: int) -> Generator[list[Doc], Any, None]:
        for batch in self.corpus_dataset.batch(batch_size):
            result: list[Doc] = []
            for id, text, embedding, tag in zip(batch["id"], batch["text"], batch["embedding"], batch["tag"]):
                result.append(
                    Doc(
                        id=id,
                        text=text,
                        embedding=np.array(embedding),
                        tag=tag,
                    )
                )
            yield result

    def corpus(self) -> Generator[Doc, Any, None]:
        for id, text, embedding, tag in zip(
            self.corpus_dataset["id"],
            self.corpus_dataset["text"],
            self.corpus_dataset["embedding"],
            self.corpus_dataset["tag"],
        ):
            yield Doc(
                id=id,
                text=text,
                embedding=np.array(embedding),
                tag=tag,
            )

    def queries(self, limit: int | None = None) -> Generator[Query, Any, None]:
        if limit is None:
            for row in self.query_dataset:
                yield Query.from_dict(row)
        else:
            for row in self.query_dataset.take(limit):
                yield Query.from_dict(row)


@dataclass
class Query:
    id: int
    text: str
    embedding: np.ndarray
    exact10: list[DocScore]
    exact90: list[DocScore]
    exact100: list[DocScore]

    @staticmethod
    def from_dict(json: dict) -> Query:
        return Query(
            id=json["id"],
            text=json["text"],
            embedding=np.array(json["embedding"]),
            exact10=[DocScore(doc, score) for doc, score in zip(json["results_10_docs"], json["results_10_scores"])],
            exact90=[DocScore(doc, score) for doc, score in zip(json["results_90_docs"], json["results_90_scores"])],
            exact100=[
                DocScore(doc, score) for doc, score in zip(json["results_100_docs"], json["results_100_scores"])
            ],
        )


@dataclass
class Doc:
    id: int
    text: str
    embedding: np.ndarray
    tag: list[int]



from hseb.core.config import Config
from hseb.engine.base import EngineBase


def test_load_class():
    config = Config.from_file("configs/nixiesearch/dev.yml")
    engine = EngineBase.load_class(name="hseb.engine.nixiesearch.NixiesearchEngine", config=config)
    assert isinstance(engine, EngineBase)



from hseb.engine.base import EngineBase
import requests
import json
import time
from hseb.core.response import DocScore, IndexResponse, SearchResponse
import tempfile
import yaml
from hseb.core.config import Config, SearchArgs, IndexArgs
from hseb.core.dataset import Doc, Query
import docker
import logging

logger = logging.getLogger()


class NixiesearchEngine(EngineBase):
    def __init__(self, config: Config):
        self.config = config

    def index_batch(self, batch: list[Doc]) -> IndexResponse:
        start = time.perf_counter()
        payload = []
        for doc in batch:
            doc_json = {
                "_id": doc.id,
                "text": {"text": doc.text, "embedding": doc.embedding.tolist()},
                "tag": doc.tag,
            }
            payload.append(doc_json)
        start = time.perf_counter()
        response = requests.post("http://localhost:8080/v1/index/test", json=payload)
        end = time.perf_counter()
        logger.debug(f"Indexed batch of {len(batch)} docs in {time.perf_counter() - start} sec")
        if response.status_code != 200:
            raise Exception(response.text)
        if "refresh_every" in self.index_args.kwargs:
            self.docs_in_segment += len(batch)
            if self.docs_in_segment >= self.index_args.kwargs["refresh_every"]:
                requests.post("http://localhost:8080/v1/index/test/flush")
                self.docs_in_segment = 0
        return IndexResponse(client_latency=end - start)

    def commit(self):
        logger.debug(requests.post("http://localhost:8080/v1/index/test/flush"))
        if self.index_args.segments is not None:
            requests.post("http://localhost:8080/v1/index/test/merge", json={"segments": self.index_args.segments})

    def index_is_green(self) -> bool:
        return True

    def search(self, search_params: SearchArgs, query: Query, top_k: int) -> SearchResponse:
        payload = {
            "query": {
                "knn": {
                    "field": "text",
                    "query_vector": query.embedding.tolist(),
                    "num_candidates": search_params.ef_search,
                    "k": top_k,
                }
            },
            "size": top_k,
        }
        if search_params.filter_selectivity != 100:
            payload["filters"] = {"include": {"term": {"tag": search_params.filter_selectivity}}}
        start = time.time_ns()

        response = requests.post("http://localhost:8080/v1/index/test/search", json=payload)
        if response.status_code != 200:
            logger.error(payload)
            raise Exception(f"got non-200 response: {response.text}")
        end = time.time_ns()
        decoded = json.loads(response.text)
        results = [DocScore(doc=int(hit["_id"]), score=float(hit["_score"])) for hit in decoded["hits"]]
        return SearchResponse(results=results, client_latency=(end - start) / 1000000000.0)

    def start(self, index_args: IndexArgs):
        self.index_args = index_args
        indexer_config = {"ram_buffer_size": index_args.kwargs.get("ram_buffer_size", "512mb")}
        if "refresh_every" in index_args.kwargs:
            indexer_config["flush"] = {"interval": "1h"}  # manual control
        if "max_merge_docs" in index_args.kwargs:
            indexer_config["merge_policy"] = {"doc_count": {"max_merge_docs": index_args.kwargs["max_merge_docs"]}}
        engine_config_file = {
            "schema": {
                "test": {
                    "config": {"indexer": indexer_config},
                    "fields": {
                        "text": {
                            "type": "text",
                            "search": {
                                "semantic": {
                                    "m": index_args.m,
                                    "ef": index_args.ef_construction,
                                    "dim": self.config.dataset.dim,
                                    "quantize": index_args.quant.value,
                                }
                            },
                        },
                        "tag": {"type": "int[]", "filter": True},
                    },
                }
            },
        }

        self.dir = tempfile.TemporaryDirectory()
        # self.container = None
        with open(f"{self.dir.name}/config.yml", "w") as config_file:
            config_file.write(yaml.safe_dump(engine_config_file))
        docker_client = docker.from_env()
        heap_size = index_args.kwargs.get("heap_size", "8g")
        self.container = docker_client.containers.run(
            image=self.config.image,
            ports={"8080/tcp": 8080},
            volumes={self.dir.name: {"bind": "/data", "mode": "rw"}},
            command="standalone -c /data/config.yml --loglevel debug",
            environment={
                "JAVA_OPTS": f"-Xmx{heap_size} -Xms{heap_size} -verbose:gc --add-modules jdk.incubator.vector"
            },
            detach=True,
        )
        self._wait_for_logs(self.container, "Ember-Server service bound to address")
        self.docs_in_segment = 0
        return self

    def stop(self, cleanup: bool):
        self.container.stop()
        self.dir.cleanup()
        if cleanup and self.container:
            self.container.remove(v=True)



import logging
import time

import docker
from opensearchpy import OpenSearch, helpers

from hseb.core.config import Config, IndexArgs, QuantDatatype, SearchArgs
from hseb.core.dataset import Doc, Query
from hseb.core.response import DocScore, IndexResponse, SearchResponse
from hseb.engine.base import EngineBase

logger = logging.getLogger()

OS_DATATYPES = {
    QuantDatatype.FLOAT32: "float",
    QuantDatatype.INT8: "byte",
    QuantDatatype.INT1: "binary",
}


class OpenSearchEngine(EngineBase):
    def __init__(self, config: Config):
        self.config = config

    def start(self, index_args: IndexArgs):
        self.index_args = index_args
        docker_client = docker.from_env()
        heap_size = index_args.kwargs.get("heap_size", "8g")
        self.container = docker_client.containers.run(
            image=self.config.image,
            ports={"9200/tcp": 9200, "9300/tcp": 9300},
            detach=True,
            environment={
                "discovery.type": "single-node",
                "OPENSEARCH_JAVA_OPTS": f"-Xms{heap_size} -Xmx{heap_size}",
                "DISABLE_SECURITY_PLUGIN": "true",
                "bootstrap.memory_lock": "true",
            },
        )
        self._wait_for_logs(self.container, "Cluster health status changed from [YELLOW] to [GREEN]")
        self.client = OpenSearch(
            hosts=[{"host": "localhost", "port": 9200}],
            http_compress=True,
            use_ssl=False,
            verify_certs=False,
            ssl_assert_hostname=False,
            ssl_show_warn=False,
            timeout=30,
        )
        index_params = {
            "knn": True,
            "number_of_shards": 1,
            "number_of_replicas": 0,
        }
        if "refresh_every" in index_args.kwargs:
            index_params["refresh_interval"] = "1h"
        if "max_merged_segment" in index_args.kwargs:
            index_params["merge"] = {"policy": {"max_merged_segment": index_args.kwargs["max_merged_segment"]}}
        field_params = {
            "type": "knn_vector",
            "dimension": self.config.dataset.dim,
            "space_type": "innerproduct",
            "data_type": "float",
            "method": {
                "name": "hnsw",
                "engine": index_args.kwargs.get("engine", "lucene"),
                "parameters": {
                    "m": index_args.m,
                    "ef_construction": index_args.ef_construction,
                },
            },
        }
        match index_args.quant:
            case QuantDatatype.FLOAT32:
                field_params["compression_level"] = "1x"
                # lucene or faiss
            case QuantDatatype.FLOAT16:
                field_params["compression_level"] = "2x"
                field_params["method"]["engine"] = "faiss"
            case QuantDatatype.INT8:
                field_params["compression_level"] = "4x"
                field_params["method"]["engine"] = "lucene"
        self.client.indices.create(
            index="test",
            body={
                "settings": {
                    "index": index_params,
                },
                "mappings": {
                    "properties": {
                        "text": field_params,
                        "tag": {"type": "integer"},
                    }
                },
            },
        )
        self.docs_in_segment = 0

    def stop(self, cleanup: bool):
        self.container.stop()
        if cleanup and self.container:
            self.container.remove(v=True)

    def commit(self):
        self.client.indices.refresh(index="test")
        if self.index_args.segments is not None:
            self.client.indices.forcemerge(
                index="test",
                params={
                    "max_num_segments": self.index_args.segments,
                    "wait_for_completion": True,
                },
            )

    def index_is_green(self) -> bool:
        response = self.client.cluster.health(index="test")
        return response["status"] == "green"

    def index_batch(self, batch: list[Doc]) -> IndexResponse:
        actions = []
        for doc in batch:
            actions.append(
                {
                    "_op_type": "index",
                    "_index": "test",
                    "_source": {
                        "text": doc.embedding.tolist(),
                        "tag": doc.tag,
                    },
                    "_id": doc.id,
                }
            )
        start = time.perf_counter()
        helpers.bulk(self.client, actions)
        end = time.perf_counter()
        self.docs_in_segment += len(batch)
        if self.docs_in_segment >= self.index_args.kwargs.get("docs_per_segment", 1024):
            self.client.indices.refresh(index="test")
            self.docs_in_segment = 0
        return IndexResponse(client_latency=end - start)

    def search(self, search_args: SearchArgs, query: Query, top_k: int) -> SearchResponse:
        knn_query = {
            "text": {
                "vector": query.embedding.tolist(),
                "k": top_k,
                "method_parameters": {
                    "ef_search": search_args.ef_search,
                },
            }
        }

        if search_args.filter_selectivity != 100:
            knn_query["text"]["filter"] = {"terms": {"tag": [search_args.filter_selectivity]}}

        search_body = {
            "size": top_k,
            "query": {"knn": knn_query},
            "_source": False,
        }

        start = time.time_ns()
        response = self.client.search(index="test", body=search_body)
        end = time.time_ns()
        return SearchResponse(
            results=[DocScore(doc=int(doc["_id"]), score=doc["_score"]) for doc in response["hits"]["hits"]],
            client_latency=(end - start) / 1000000000.0,
        )



from hseb.core.config import (
    QuantDatatype,
)

from tests.engine.base import EngineParams, EngineSuite


class TestQdrantEngine(EngineSuite):
    def params(self):
        return EngineParams(
            engine="hseb.engine.qdrant.Qdrant",
            image="qdrant/qdrant:v1.15.4",
            quantizations=[QuantDatatype.FLOAT16, QuantDatatype.INT8, QuantDatatype.INT1],
        )



from abc import ABC, abstractmethod
import time

from hseb.core.config import Config, DatasetConfig, ExperimentConfig, IndexArgsMatrix, QuantDatatype, SearchArgsMatrix
from hseb.core.dataset import BenchmarkDataset, Doc
from hseb.core.measurement import ExperimentResult, QueryResult
from hseb.core.submission import ExperimentMetrics
from hseb.engine.base import EngineBase
from tqdm import tqdm
from structlog import get_logger
from dataclasses import dataclass, field

logger = get_logger()


@dataclass
class EngineParams:
    engine: str
    image: str
    quantizations: list[QuantDatatype] = field(default_factory=list)
    index_kwargs: dict[str, list] = field(default_factory=dict)


class EngineSuite(ABC):
    @abstractmethod
    def params(self) -> EngineParams: ...

    def config(self) -> Config:
        engine_params = self.params()
        return Config(
            engine=engine_params.engine,
            image=engine_params.image,
            dataset=DatasetConfig(
                dim=384,
                name="hseb-benchmark/msmarco",
                query="query-all-MiniLM-L6-v2-1K",
                corpus="corpus-all-MiniLM-L6-v2-1K",
            ),
            experiments=[
                ExperimentConfig(
                    tag="test",
                    k=10,
                    index=IndexArgsMatrix(
                        m=[16],
                        ef_construction=[64],
                        quant=[QuantDatatype.FLOAT32],
                        kwargs=engine_params.index_kwargs,
                    ),
                    search=SearchArgsMatrix(ef_search=[16], filter_selectivity=[10, 100]),
                ),
                ExperimentConfig(
                    tag="test-quant",
                    k=10,
                    index=IndexArgsMatrix(
                        m=[16],
                        ef_construction=[64],
                        quant=engine_params.quantizations,
                        kwargs=engine_params.index_kwargs,
                    ),
                    search=SearchArgsMatrix(ef_search=[16], filter_selectivity=[100]),
                ),
            ],
        )

    def test_index_search(self):
        conf: Config = self.config()
        data = BenchmarkDataset(conf.dataset)
        engine = EngineBase.load_class(conf.engine, config=conf)

        docs: dict[int, Doc] = {doc.id: doc for doc in data.corpus()}

        for exp in conf.experiments:
            for index_args in exp.index.expand():
                try:
                    logger.info("starting engine")
                    engine.start(index_args)
                    logger.info(f"indexing: {index_args}")
                    for batch in data.corpus_batched(index_args.batch_size):
                        engine.index_batch(batch)

                    engine.commit()
                    is_green = False
                    attempts = 0
                    while not is_green and attempts < 30:
                        is_green = engine.index_is_green()
                        time.sleep(1)
                    assert is_green
                    for search_args in exp.search.expand():
                        measurements = []
                        logger.info(f"searching: {search_args}")
                        for query in tqdm(list(data.queries(limit=10)), desc="searching"):
                            results = engine.search(search_args, query, 16)
                            assert len(results.results) == 16
                            measurements.append(QueryResult.from_response(query, search_args, results))
                            prev_score = 10000.0
                            for doc in results.results:
                                assert doc.score <= prev_score
                                assert isinstance(doc.doc, int)
                                real_doc = docs[doc.doc]
                                assert search_args.filter_selectivity in real_doc.tag
                                prev_score = doc.score
                        result = ExperimentResult(
                            tag="test",
                            indexing_time=[1],
                            index_args=index_args,
                            search_args=search_args,
                            queries=measurements,
                            warmup_latencies=[1],
                        )
                        metrics = ExperimentMetrics.from_experiment(result)
                        recall10 = sum(metrics.metrics.recall10) / len(metrics.metrics.recall10)
                        assert recall10 > 0.01
                finally:
                    engine.stop(cleanup=True)



import logging
import time
from uuid import UUID
import uuid

import docker
import weaviate
from weaviate.classes.config import Configure, VectorDistances, DataType, Property
from weaviate.classes.query import Filter, MetadataQuery
from weaviate.classes.data import DataObject


from hseb.core.config import Config, IndexArgs, QuantDatatype, SearchArgs
from hseb.core.dataset import Doc, Query
from hseb.core.response import DocScore, IndexResponse, SearchResponse
from hseb.engine.base import EngineBase

logger = logging.getLogger()

WEAVIATE_DATATYPES = {
    QuantDatatype.FLOAT32: None,  # Default precision
    QuantDatatype.INT8: Configure.VectorIndex.Quantizer.sq(),  # Scalar quantization for int8
    QuantDatatype.INT1: Configure.VectorIndex.Quantizer.bq(),  # Binary quantization for int1
}


class WeaviateEngine(EngineBase):
    def __init__(self, config: Config):
        self.config = config
        self.client = None
        self.container = None
        self.collection_name = "test"

    def start(self, index_args: IndexArgs):
        if index_args.segments is not None:
            raise ValueError("Weaviate cannot set number of segments")

        docker_client = docker.from_env()

        self.container = docker_client.containers.run(
            image=self.config.image,
            ports={"8080/tcp": 8080, "50051/tcp": 50051},
            detach=True,
            environment={"PERSISTENCE_DATA_PATH": "/var/lib/weaviate"},
        )
        self._wait_for_logs(self.container, "Serving weaviate at")

        # Wait a bit more to ensure Weaviate is fully ready
        time.sleep(3)

        # Connect to Weaviate
        self.client = weaviate.connect_to_local(port=8080)

        # Configure quantizer based on quantization type
        quantizer = WEAVIATE_DATATYPES.get(index_args.quant, None)

        # Create collection (using deprecated parameter for now to test basic functionality)
        self.client.collections.create(
            name=self.collection_name,
            vector_config=Configure.Vectors.self_provided(
                name="self",
                vector_index_config=Configure.VectorIndex.hnsw(
                    distance_metric=VectorDistances.DOT,
                    ef_construction=index_args.ef_construction,
                    max_connections=index_args.m,
                ),
                quantizer=quantizer,
            ),
            properties=[
                Property(name="text", data_type=DataType.TEXT),
                Property(name="tag", data_type=DataType.INT_ARRAY, index_filterable=True),
            ],
        )

        self.index_args = index_args
        self.uuid_docid_cache: dict[UUID, int] = {}
        return self

    def stop(self, cleanup: bool):
        if self.client:
            self.client.close()
        if self.container:
            self.container.stop()
        if cleanup and self.container:
            self.container.remove(v=True)

    def commit(self):
        pass

    def index_is_green(self) -> bool:
        all_shards_green = True
        for shard in self.client.collections.get(self.collection_name).shards():
            if shard.vector_indexing_status != "READY":
                all_shards_green = False
        return all_shards_green

    def index_batch(self, batch: list[Doc]) -> IndexResponse:
        collection = self.client.collections.get(self.collection_name)

        # Prepare objects for batch insertion
        objects = []
        for doc in batch:
            key = uuid.uuid4()
            self.uuid_docid_cache[key] = doc.id
            obj = DataObject(
                properties={
                    "text": doc.text,
                    "tag": doc.tag,
                },
                uuid=key,
                vector=doc.embedding.tolist(),
            )
            objects.append(obj)

        # Batch insert
        start = time.perf_counter()
        collection.data.insert_many(objects)
        end = time.perf_counter()
        return IndexResponse(client_latency=end - start)

    def search(self, search_params: SearchArgs, query: Query, top_k: int) -> SearchResponse:
        collection = self.client.collections.get(self.collection_name)

        # Create filter if needed
        filters = None
        if search_params.filter_selectivity != 100:
            filters = Filter.by_property("tag").equal(search_params.filter_selectivity)

        start = time.time_ns()

        # Perform vector search with dot product similarity
        response = collection.query.near_vector(
            near_vector=query.embedding.tolist(),
            limit=top_k,
            filters=filters,
            return_metadata=MetadataQuery(distance=True),
        )

        end = time.time_ns()

        # Convert results to DocScore objects
        doc_scores = []
        for obj in response.objects:
            id = self.uuid_docid_cache[obj.uuid]
            doc_scores.append(DocScore(doc=id, score=-obj.metadata.distance))

        return SearchResponse(
            results=doc_scores,
            client_latency=(end - start) / 1000000000.0,
        )



from hseb.core.dataset import BenchmarkDataset
from hseb.core.config import DatasetConfig


class TestDataset:
    def test_dataset_loading(self):
        conf = DatasetConfig(
            dim=384,
            name="hseb-benchmark/msmarco",
            query="query-all-MiniLM-L6-v2-1K",
            corpus="corpus-all-MiniLM-L6-v2-1K",
        )
        ds = BenchmarkDataset(conf)
        assert len(ds.corpus_dataset) == 1000
        assert len(list(ds.corpus_batched(100))) == 10
        assert len(ds.query_dataset) == 10000
        assert len(list(ds.queries())) == 10000



import argparse
from datasets import load_dataset, Features, Value, Dataset
from typing import Dict, List
from sentence_transformers import SentenceTransformer
import numpy as np
from functools import partial
import faiss
import os
from tqdm import tqdm

selectivities = [10, 90, 100]


def add_index_tags(batch: Dict[str, List], indices: List[int]) -> Dict[str, List]:
    tags = []
    for index in indices:
        doc_tags = []
        for s in selectivities:
            if index % 100 <= s:
                doc_tags.append(s)
        tags.append(doc_tags)
    return {"id": indices, "tag": tags}


def add_index(batch: Dict[str, List], indices: List[int]) -> Dict[str, List]:
    return {"id": indices}


def zip_embed(embeds: np.ndarray, batch: Dict[str, List], indices: List[int]) -> Dict[str, List]:
    return {"embedding": [embeds[i] for i in indices]}


def zip_ground_truth(
    docs: np.ndarray, scores: np.ndarray, selectivity: int, batch: dict[str, list], indices: list[int]
) -> dict[str, list]:
    return {
        f"results_{selectivity}_docs": [docs[i] for i in indices],
        f"results_{selectivity}_scores": [scores[i] for i in indices],
    }


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--queries", required=True, type=str, help="path to queries.json")
    parser.add_argument(
        "--queries-sample", required=False, type=int, default=10000, help="number of queries to randomly sample"
    )
    parser.add_argument("--corpus", required=True, type=str, help="path to corpus.json")
    parser.add_argument(
        "--corpus-sample", required=False, type=int, default=100000, help="number of documents to randomly sample"
    )
    parser.add_argument("--model", type=str, required=True, help="SBERT embedding model to use")
    parser.add_argument("--batch-size", type=int, required=False, default=512)
    parser.add_argument("--top-n", type=int, required=False, default=100)
    parser.add_argument("--out", type=str, required=False, default="out")

    args = parser.parse_args()

    schema = Features({"text": Value("string")})
    queries: Dataset = load_dataset("json", data_files={"train": args.queries}, split="train").select_columns("text")
    if args.queries_sample:
        queries = queries.shuffle().take(args.queries_sample)

    queries = queries.map(
        function=add_index,
        with_indices=True,
        batched=True,
        desc="indexing queries",
    )
    corpus: Dataset = load_dataset("json", data_files={"train": args.corpus}, split="train").select_columns("text")
    if args.corpus_sample:
        corpus = corpus.shuffle().take(args.corpus_sample)
    corpus = corpus.map(
        function=add_index_tags,
        with_indices=True,
        batched=True,
        desc="adding selectivity tags",
    )

    model = SentenceTransformer(args.model)
    print("encoding queries")
    query_embeds = model.encode_query(
        sentences=list(queries.to_dict()["text"]),
        convert_to_numpy=True,
        show_progress_bar=True,
        batch_size=args.batch_size,
        normalize_embeddings=True,
        # device=["cuda:0", "cuda:1"],
    )
    queries = queries.map(
        function=partial(zip_embed, query_embeds),
        with_indices=True,
        batched=True,
        desc="joining query embeddings",
    )
    print("encoding docs")
    corpus_embeds = model.encode_document(
        sentences=list(corpus.to_dict()["text"]),
        convert_to_numpy=True,
        show_progress_bar=True,
        batch_size=args.batch_size,
        normalize_embeddings=True,
        # device=["cuda:0", "cuda:1"],
    )
    corpus = corpus.map(
        function=partial(zip_embed, corpus_embeds),
        with_indices=True,
        batched=True,
        desc="joining doc embeddings",
    )

    print("Building Flat index")
    faiss.omp_set_num_threads(os.cpu_count())
    index = faiss.IndexFlatIP(corpus_embeds.shape[1])
    index.add(corpus_embeds)
    print("Exact search...")
    size = corpus_embeds.shape[0]
    docs: Dict[int, np.ndarray] = {}
    doc_scores: Dict[int, np.ndarray] = {}
    for selectivity in tqdm(selectivities):
        bitmap = np.zeros(size, dtype=bool)
        targets = [index for index in range(size) if index % 100 <= selectivity]
        bitmap[targets] = True
        packed = np.packbits(bitmap, bitorder="little")
        params = faiss.SearchParameters(sel=faiss.IDSelectorBitmap(packed))
        scores, docs = index.search(query_embeds, args.top_n, params=params)
        queries = queries.map(
            function=partial(zip_ground_truth, docs, scores, selectivity),
            with_indices=True,
            batched=True,
            desc="zipping",
        )

    corpus.to_json(f"{args.out}/{args.model}/{args.corpus_sample}/corpus.jsonl")
    queries.to_json(f"{args.out}/{args.model}/{args.corpus_sample}/queries.jsonl")



from hseb.core.config import (
    QuantDatatype,
)

from tests.engine.base import EngineParams, EngineSuite


class TestPostgresEngine(EngineSuite):
    def params(self):
        return EngineParams(
            engine="hseb.engine.postgres.PostgresEngine",
            image="pgvector/pgvector:0.8.1-pg17-trixie",
            quantizations=[QuantDatatype.FLOAT16],
        )



import logging
import time

import docker
import redis
from redis.commands.search.query import Query as RedisQuery
from redis.commands.search.field import VectorField, TagField
from redis.commands.search.index_definition import IndexDefinition, IndexType

from hseb.core.config import Config, IndexArgs, QuantDatatype, SearchArgs
from hseb.core.dataset import Doc, Query
from hseb.core.response import DocScore, IndexResponse, SearchResponse
from hseb.engine.base import EngineBase

logger = logging.getLogger()

REDIS_DATATYPES = {
    QuantDatatype.FLOAT32: "FLOAT32",
    QuantDatatype.FLOAT16: "FLOAT16",
    QuantDatatype.INT8: "INT8",
}


class RedisEngine(EngineBase):
    def __init__(self, config: Config):
        self.config = config
        self.client = None
        self.container = None

    def start(self, index_args: IndexArgs):
        # Check for unsupported quantization types
        if index_args.quant not in REDIS_DATATYPES:
            raise ValueError(
                f"Redis does not support {index_args.quant} quantization. Supported types: {list(REDIS_DATATYPES.keys())}"
            )
        if index_args.segments is not None:
            raise ValueError("Redis cannot set number of segments")

        docker_client = docker.from_env()

        # Redis configuration
        maxmemory = index_args.kwargs.get("maxmemory", "2gb")
        maxmemory_policy = index_args.kwargs.get("maxmemory_policy", "allkeys-lru")

        self.container = docker_client.containers.run(
            image=self.config.image,
            ports={"6379/tcp": 6379},
            detach=True,
            command=["redis-server", "--maxmemory", maxmemory, "--maxmemory-policy", maxmemory_policy],
        )
        self._wait_for_logs(self.container, "Ready to accept connections")

        # Wait a bit more to ensure Redis is fully ready
        time.sleep(2)

        # Connect to Redis
        self.client = redis.Redis(host="localhost", port=6379, decode_responses=False)

        # Create index with vector field using direct FT.CREATE command

        self.client.ft("documents").create_index(
            fields=[
                VectorField(
                    name="embedding",
                    algorithm="HNSW",
                    attributes={
                        "TYPE": REDIS_DATATYPES[index_args.quant],
                        "DIM": self.config.dataset.dim,
                        "DISTANCE_METRIC": "IP",
                        "M": index_args.m,
                        "EF_CONSTRUCTION": index_args.ef_construction,
                    },
                ),
                TagField("tag", separator=","),  # Use comma separator for tags
            ],
            definition=IndexDefinition(prefix=["doc:"], index_type=IndexType.HASH),
        )

        self.index_args = index_args
        return self

    def stop(self, cleanup: bool):
        if self.client:
            self.client.close()
        if self.container:
            self.container.stop()
        if cleanup and self.container:
            self.container.remove(v=True)

    def commit(self):
        pass

    def index_is_green(self) -> bool:
        info = self.client.ft("documents").info()
        return info["indexing"] == 0

    def index_batch(self, batch: list[Doc]) -> IndexResponse:
        pipe = self.client.pipeline()

        for doc in batch:
            # Use the same dtype as the index was created with
            match self.index_args.quant:
                case QuantDatatype.FLOAT16:
                    embedding_bytes = doc.embedding.astype("float16").tobytes()
                case QuantDatatype.FLOAT32:
                    embedding_bytes = doc.embedding.astype("float32").tobytes()
                case QuantDatatype.INT8:
                    embedding_bytes = doc.embedding.astype("int8").tobytes()

            pipe.hset(
                f"doc:{doc.id}",
                mapping={
                    "embedding": embedding_bytes,
                    "tag": ",".join(map(str, doc.tag)),  # Store tags as comma-separated string
                },
            )
        start = time.perf_counter()
        pipe.execute()
        end = time.perf_counter()
        return IndexResponse(client_latency=end - start)

    def search(self, search_params: SearchArgs, query: Query, top_k: int) -> SearchResponse:
        # Build KNN query - use the same dtype as the index was created with
        match self.index_args.quant:
            case QuantDatatype.FLOAT16:
                embedding_bytes = query.embedding.astype("float16").tobytes()
            case QuantDatatype.FLOAT32:
                embedding_bytes = query.embedding.astype("float32").tobytes()
            case QuantDatatype.INT8:
                embedding_bytes = query.embedding.astype("int8").tobytes()

        # Create base query with AS clause for scoring
        base_query = f"(*)=>[KNN {top_k} @embedding $query_vector AS vector_score]"

        # Add filter if needed
        if search_params.filter_selectivity != 100:
            # Filter by tag using TagField exact match syntax with curly brackets
            base_query = f"(@tag:{{{search_params.filter_selectivity}}})=>[KNN {top_k} @embedding $query_vector AS vector_score]"

        # Use ef_search parameter for HNSW search

        start = time.time_ns()

        results = self.client.ft("documents").search(
            RedisQuery(base_query)
            .sort_by("vector_score", asc=True)
            .return_fields("vector_score")
            .dialect(2)
            .paging(0, top_k),
            query_params={"query_vector": embedding_bytes},
        )
        end = time.time_ns()

        # Convert results to DocScore objects
        doc_scores = []
        for result in results.docs:
            # Extract document ID from key (doc:123 -> 123)
            doc_id = int(result.id.split(":")[1])
            # Redis returns similarity score in vector_score field
            score = float(result.vector_score)
            doc_scores.append(DocScore(doc=doc_id, score=-score))

        return SearchResponse(
            results=doc_scores,
            client_latency=(end - start) / 1000000000.0,
        )



import argparse
from hseb.core.report import Report
from hseb.core.submission import Submission
from structlog import get_logger

logger = get_logger()

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", type=str, required=True, help="Path to a submission file")
    parser.add_argument("--report", type=str, required=True, help="Name of report file to write")
    args = parser.parse_args()
    logger.info(f"Writing report: {args}")

    submission = Submission.from_json(args.input)
    report = Report.from_experiments(submission.experiments)
    report.df.to_json(args.report)
    logger.info(f"Report written to {args.report}")



from hseb.core.config import Config, IndexArgsMatrix, SearchArgsMatrix


class TestConfig:
    def test_loading(self):
        yaml_content = """
engine: nixiesearch
image: nixiesearch/nixiesearch:0.6.3
dataset: 
  dim: 384
  name: hseb-benchmark/msmarco
  query: "query-all-MiniLM-L6-v2-1K"
  corpus: "corpus-all-MiniLM-L6-v2-1K"
batch_size: 1024
experiments:
  - tag: test
    k: 10
    index:
      quant: [float32]
      m: [16]
      ef_construction: [64]
    search:
      ef_search: [16,32,64]
      filter_selectivity: [10, 90, 100]
"""
        loaded = Config.from_yaml(yaml_content)
        assert loaded.engine == "nixiesearch"

    def test_index_args_expand(self):
        args = IndexArgsMatrix(m=[1, 2], ef_construction=[3, 4], quant=["float32"], kwargs={"foo": [5, 6]})
        expanded = args.expand()
        assert len(expanded) == 8

    def test_search_args_expand(self):
        args = SearchArgsMatrix(ef_search=[1, 2], filter_selectivity=[3, 4], kwargs={"foo": [5, 6]})
        expanded = args.expand()
        assert len(expanded) == 8



from __future__ import annotations
from typing import Type
from hseb.core.response import IndexResponse, SearchResponse
import time
from abc import ABC, abstractmethod
from hseb.core.dataset import Query, Doc
from hseb.core.config import Config, SearchArgs, IndexArgs
import logging
import importlib

logger = logging.getLogger()


class EngineBase(ABC):
    @abstractmethod
    def start(self, index_args: IndexArgs) -> None: ...

    @abstractmethod
    def stop(self, cleanup: bool) -> None: ...

    @abstractmethod
    def commit(self) -> None: ...

    @abstractmethod
    def index_is_green(self) -> bool: ...

    @abstractmethod
    def index_batch(self, batch: list[Doc]) -> IndexResponse: ...

    @abstractmethod
    def search(self, search_args: SearchArgs, query: Query, top_k: int) -> SearchResponse: ...

    def _wait_for_logs(self, container, log_message: str, timeout: int = 60):
        """Wait for a specific log message to appear in container logs"""
        start_time = time.time()
        while time.time() - start_time < timeout:
            logs = container.logs().decode("utf-8")
            # print(logs)
            if log_message in logs:
                return True
            time.sleep(1)
        raise TimeoutError(f"Timeout waiting for log message: {log_message}")

    @staticmethod
    def load_class(name: str, config: Config) -> EngineBase:
        module_name, class_name = name.rsplit(".", 1)
        logger.debug(f"Loading model {class_name} from module {module_name}")
        module = importlib.import_module(module_name)
        if not hasattr(module, class_name):
            raise ValueError(f"Cannot find {class_name} in module {module_name}")
        cls: Type[EngineBase] = getattr(module, class_name)
        if not issubclass(cls, EngineBase):
            raise ValueError(f"Class {class_name} is not a EngineBase")

        obj = cls(config=config)
        return obj



from hseb.core.config import (
    QuantDatatype,
)

from tests.engine.base import EngineParams, EngineSuite


class TestOpensearch3Engine(EngineSuite):
    def params(self):
        return EngineParams(
            engine="hseb.engine.opensearch.OpenSearchEngine",
            image="opensearchproject/opensearch:3.2.0",
            quantizations=[QuantDatatype.INT8, QuantDatatype.FLOAT16],
            index_kwargs={"engine": ["lucene", "faiss"]},
        )



import time
from hseb.core.config import (
    Config,
    DatasetConfig,
    ExperimentConfig,
    IndexArgs,
    IndexArgsMatrix,
    QuantDatatype,
    SearchArgs,
    SearchArgsMatrix,
)
from hseb.core.measurement import ExperimentResult, QueryResult
from hseb.core.response import DocScore
import tempfile
import os

from hseb.core.submission import ExperimentMetrics, Submission


def test_submission():
    config = Config(
        engine="hseb.engine.nixiesearch.nixiesearch.NixiesearchEngine",
        image="nixiesearch/nixiesearch:0.6.5",
        dataset=DatasetConfig(
            dim=384,
            name="hseb-benchmark/msmarco",
            query="query-all-MiniLM-L6-v2-1K",
            corpus="corpus-all-MiniLM-L6-v2-1K",
        ),
        experiments=[
            ExperimentConfig(
                tag="test",
                k=10,
                index=IndexArgsMatrix(m=[16], ef_construction=[64], quant=["float32"]),
                search=SearchArgsMatrix(ef_search=[16], filter_selectivity=[100]),
            )
        ],
    )
    m1 = QueryResult(query_id=1, exact=[DocScore(1, 1)], response=[DocScore(1, 1)], client_latency=1)
    m2 = QueryResult(query_id=1, exact=[DocScore(1, 1)], response=[DocScore(1, 1)], client_latency=1)
    result = ExperimentResult(
        tag="test",
        index_args=IndexArgs(
            m=32,
            ef_construction=32,
            quant=QuantDatatype.FLOAT32,
            batch_size=32,
            segments=None,
            kwargs={},
        ),
        search_args=SearchArgs(ef_search=32, filter_selectivity=100, kwargs={}),
        queries=[m1, m2],
        indexing_time=[1],
        warmup_latencies=[1],
    )
    with tempfile.TemporaryDirectory(prefix="hseb_test_") as dir:
        result.to_json(dir)
        sub = Submission.from_dir(config=config, path=dir)
        assert len(sub.experiments) == 1
        assert len(sub.experiments[0].metrics.latency) == 2
        export_path = os.path.join(tempfile.tempdir, f"hseb_export_test_{time.time()}.json")

        sub.to_json(export_path)
        decoded = Submission.from_json(export_path)
        assert decoded == sub


def test_recall():
    r1 = ExperimentMetrics.recall_score([1, 2, 3, 4], [1, 2, 4, 5], 4)
    assert r1 == 0.75
